from marquetry.datasets import csv_loader
from marquetry.datasets import fashion_mnist
from marquetry.datasets import mnist
from marquetry.datasets import trigonometric_curve
from marquetry.datasets import spiral
from marquetry.datasets import titanic

from marquetry.datasets.csv_loader import CsvLoader
from marquetry.datasets.fashion_mnist import FashionMNIST
from marquetry.datasets.mnist import MNIST
from marquetry.datasets.trigonometric_curve import TrigonometricCurve
from marquetry.datasets.spiral import Spiral
from marquetry.datasets.titanic import Titanic
